# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this file,
# You can obtain one at http://mozilla.org/MPL/2.0/.
"""
Helper APIs for accessing various storage services.

This sub-package provides helper APIs for accessing various backend storage
services.  Currently available are:

    * mozsvc.storage.mcclient:  client for interacting with memcache

More may be added in the future, e.g. an SQL database access layer.
"""
